'******************************************************************************
' C Function Library in Spin
' Copyright (c) 2010, 2011, 2012, Dave Hein
' See end of file for terms of use.
'******************************************************************************
'******************************************************************************
' Revison History
' v1.0 - 4/2/2010 First official release
'******************************************************************************
{{ 
  This object provides some standard C library functions written in Spin.
  There are three main types of functions provided -- string manipulation,
  formatted I/O and memory allocation.  C functions that use a variable
  number of arguments are supported in two ways.  The first way uses
  unique function names for each number of arguments that is used, such
  as printf3 for a printf with three arguments.  The second way uses an
  argument list, such as vprintf.  The argument list consists of an array
  of longs that holds the values of the arguments.

  The start function must be called before using any of the serial I/O or
  memory allocation functions.  It will setup a a serial port that transmits on
  pin 30 and receives on pin 31 at 57600 baud.  It will also establish a stack
  space of 200 longs for the top object.  The malloc heap begins immediately
  after the stack space, and extends to the end of the RAM.

  Serial I/O and stack space parameters can be specified by calling start1
  instead of start.  The first four parameters of start1 are the same as those
  used in a call to the FullDuplexSerial start function.  The fifth parameter,
  stacksize, defines the size in longs of the stack space.

  Additional serial ports can be created by calling the openserial routine.
  In additional to rxpin, txpin, mode and baudrate the receive and transmit
  buffer sizes are also specified.  A fileinfo pointer is returned, which
  is used when calling fputc, fgetc and all of the other f... routines.

  The functions puts, putchar and getchar use the str1, tx1 and rx1 methods
  of a modified FullDuplexSerial object.  This object, cserial, allows for multiple
  serial ports through the use of a structer pointer called a handle.  See
  cserial.spin for more information.

  clib uses a file info structure to access I/O devices, such as the serial port.
  The I/O routines that start with the letter "f" require a pointer to the file info
  struction, wich is normally named pfileinfo.  Currently, only serial I/O and
  string I/O are supported.  Future releases will also support file I/O.
}}

OBJ
  mem : "cmalloc"
  fileio : "fsrwx"
  sys : "sysdefs"
  'ldr : "loader"
  ser : "cserial"

CON
  NEW_LINE = 10
  SEEK_SET = 0
  SEEK_CUR = 1
  SEEK_END = 2

  file_info_fileio = 2
  
  file_info_type   = 0
  file_info_data   = 1
  file_info_size   = 8

VAR
  long firstvar

DAT
  currentwd long 0[25]
  tempwd long 0[25]
  
' A process jumps here when it does the final return
' This must be the first PUB method so we can locate it in the table
PUB exit | locknum, cognum
  'ser.str(string("exit returned "))
  'ser.dec(long[@result-64-8])
  'ser.tx(13)
  locknum := long[sys#filelock]
  repeat until not lockset(locknum - 1)
  cognum := cogid
  long[sys#fileparm] := @cognum
  long[sys#filecmd] := (cogid << 16) | 25 ' kill command 
  repeat while long[sys#filecmd]
  ' Should never get to this point
  lockclr(locknum - 1)
  'ser.str(string("Exit failed", 13))
  repeat

'******************************************************************************  
' String Routines
'******************************************************************************  
PRI strcpy(dst, src) | slen
{{
  This routine copies the string pointed to by "src" to the location pointed
  to by "dst".  It returns the value of "dst".
}}
  slen := strsize(src) + 1
  bytemove(dst, src, slen)
  return dst

PRI strncpy(dst, src, num) | slen
{{
  This routine copies the first num bytes from the src string to the dst string.
  If src contains less than num bytes, then only the bytes contained in the src
  string are copied, and the remaining bytes are set to zero.  Note that if num
  is less than or equal to the string length of src the dst string will not be
  terminated with a NULL.  This routine returns the value of the dst pointer.
}}
  if (num < 1)
    return dst
  slen := strsize(src)
  if (slen > num)
    slen := num
  bytemove(dst, src, slen)
  bytefill(dst + slen, 0, num - slen)
  return dst

PUB fgetc(pfileinfo) | char, handle, num
{{
  This routine returns a single character from the input device defined by "pfileinfo".
  It will not return until a character has been received.
}}
  char := 0
  handle := long[pfileinfo][file_info_data]
  num := fileio.hread(handle, @char, 1)
  if (num < 1)
    return -1
  return char

PUB fgets(str, num, pfileinfo) | handle, i, value
{{
  This routine returns a string from the input device defined by "pfileinfo".  It will
  return until it has either received a carriage return or "size" number of characters.
}}
  handle := long[pfileinfo][file_info_data]
  result := str
  repeat i from 0 to num - 2
    value := fileio.hread(handle, str, 1)
    IF (value < 1)
      IF (i == 0)
        RETURN 0
      QUIT
    IF (BYTE[str] == 0)
      QUIT
    if byte[str++] == NEW_LINE
      quit
  BYTE[str] := 0

PRI ProcessPath(fname, pathstr, filestr) | fname1
  strncpy(pathstr, fname, 80)
  fname1 := SplitPathFile(pathstr)
  if byte[fname1] == 0 AND byte[pathstr] == "/"
    'ser.dbprintf0(string("FIXIT\n"))
    fname1 := filestr
    strcpy(fname1, pathstr + 1)
    byte[pathstr][1] := 0

  'ser.dbprintf3(string("fopen(%s %s %s)\n"), @pathstr, fname1, mode)
  if fname1 <> pathstr
    strcpy(@tempwd, @currentwd)
    chdir(pathstr)

  return fname1

PUB fopen(fname, mode) | pfileinfo, file_data, handlesize, value, pathstr[20], fname1, filestr[8]
{{
  This routine opens a file with the name pointed to by "fname".  The mode is
  determined by the first character of the string pointed to by "mode".  The first
  character must be "r", "w" or "a" for read, write or append.
  It returns a pointer to the file info structure if successful, or a NULL value
  if not successful.
}}
  fname1 := ProcessPath(fname, @pathstr, @filestr)

  'ser.dbprintf2(string("fopen %s %s\n"), fname, mode)
  ' Change directory if directory path is specified
  handlesize := fileio.handlesize

  'ser.dbprintf1(string("fopen: handlesize = %d\n"), handlesize)
  ifnot (pfileinfo := mem.malloc(file_info_size + handlesize))
    ChangeTempWD
    return 0
  file_data := pfileinfo + file_info_size
  long[pfileinfo][file_info_data] := file_data

  long[pfileinfo][file_info_type] := file_info_fileio
  value := fileio.hopen(file_data, fname1, byte[mode])

  ChangeTempWD
  if value
    mem.free(pfileinfo)
    return 0
  'ser.dbprintf0(string("SUCCESS\n"))
  return pfileinfo

PRI ChangeTempWD
  if byte[@tempwd]
    chdir(@tempwd)
    byte[@tempwd] := 0

PRI SplitPathFile(pathstr) | fname, len, len1
  ' Search backwards for a "/"
  len1 := len := strsize(pathstr)
  ifnot len
    return pathstr
  if len == 1 and byte[pathstr] == "/"
    return pathstr + 1
  repeat while --len > 0
    if byte[pathstr][len] == "/"
      byte[pathstr][len] := 0
      return pathstr + len + 1
  if byte[pathstr] == "/"
    return pathstr + len1
  else
    return pathstr

PUB fread(buffer, elem_size, num, pfileinfo) | file_data
{{
  This routine reads a block of data from the input device defined by "pfileinfo".
  It returns the number of bytes read, or a -1 if the end of file is
  encountered.
}}
  file_data := long[pfileinfo][file_info_data]
  result := fileio.hread(file_data, buffer, elem_size * num)
  
PUB fwrite(buffer, elem_size, num, pfileinfo) | file_data
{{
  This routine writes a block of data to the output device defined by "pfileinfo".
  It returns the number of bytes written, or a -1 if there is an error.
}}
  file_data := long[pfileinfo][file_info_data]
  result := fileio.hwrite(file_data, buffer, elem_size * num)

PUB fclose(pfileinfo) | file_data
  file_data := long[pfileinfo][file_info_data]
  result := fileio.hclose(file_data)
  ifnot result
    result := 1
  mem.free(pfileinfo)

PUB remove(fname) | pathstr[20], fname1, filestr[8]
  'printf1(string("remove %s\n"), fname)
  fname1 := ProcessPath(fname, @pathstr, @filestr)
  result := fileio.hopen(0, fname1, "d")
  ChangeTempWD
  ifnot result
    result := 1
    
PUB rename(fname1, fname2)
  'printf1(string("remove %s\n"), fname)
  result := fileio.rename(fname1, fname2)   

PUB create(fname) | handle, pathstr[20], fname1, filestr[8]
  'printf2(string("create: %s %d\n"), fname, currentdisk)
  'handle := mem.malloc(fileio.handlesize)
  'printf2(string("create1: %s $%04x\n"), fname, handle)
  ifnot handle
    return -1
  fname1 := ProcessPath(fname, @pathstr, @filestr)
  ifnot (handle := fileio.hopen(handle, fname1, "r"))
    fileio.hclose(handle)
    result := 1
  elseif result == -1
    ifnot (handle := fileio.hopen(handle, fname1, "w"))
      fileio.hclose(handle)
      result := 1
  'mem.free(handle)
  ChangeTempWD

PUB getfilesize(pfileinfo) | file_data
  file_data := long[pfileinfo][file_info_data]
  result := fileio.hget_filesize(file_data)

PUB mount_explicit(DO, CLK, DI, CS)
  result := fileio.mount_explicit(DO, CLK, DI, CS)
  bytemove(@currentwd, string("/"), 2)
  ifnot result
    result := 1

PUB getcwd(str, num)
  strncpy(str, @currentwd, num)

PUB FindChar(str, value_0)
  REPEAT WHILE (BYTE[str])
    IF (BYTE[str] == value_0)
      QUIT
    str++
  RETURN str

PUB chdir(path) | abspath[25], cwdlen, abslen, fname
  'ser.dbprintf1(string("chdir(%s)\n"), path)

  ' Check if current directory
  if byte[path] == 0 OR strcomp(path, string("."))
    return

  ' Check for absolute path
  if byte[path] == "/"
    strcpy(@abspath, path)
    abslen := strsize(@abspath)
    cwdlen := strsize(@currentwd)
    path := @abspath + cwdlen
    ' Check if relative from CWD
    if abslen > cwdlen AND byte[path] == "/"
      byte[path] := 0
      if strcomp(@currentwd, @abspath)
        byte[path++] := "/"
      else
        byte[path] := "/"
        path := @abspath
    else
      path := @abspath

  ' Check if going up one level
  elseif strcomp(path, string(".."))
    path := @abspath
    strcpy(path, @currentwd)
    fname := SplitPathFile(path)
    if byte[fname] == 0 AND byte[path] == "/"
      strcpy(path, string("/"))

  ' Else, relative path to current working directory
  else
    strcpy(@abspath, @currentwd)
    abslen := strsize(@abspath)
    if byte[@abspath + abslen - 1] <> "/"
      byte[@abspath + abslen++] := "/"
    strcpy(@abspath + abslen, path)
    path := @abspath + abslen

  'ser.dbprintf1(string("chdir: expanded path = %s\n"), @abspath)

  ifnot (result := fileio.chdir(path))
    strcpy(@currentwd, @abspath)

'PUB format(size)
'  fileio.format(size)

PUB opendir(path) | handlesize, pfileinfo, file_data, value
  strcpy(@tempwd, @currentwd)
  'ser.dbprintf1(string("opendir %s\r\n"), path)
  if chdir(path)
    'ser.dbprintf0(string("chdir failed\r\n"))
    ChangeTempWD
    return 0
  handlesize := fileio.handlesize
  'ser.dbprintf1(string("fopen: handlesize = %d\n"), handlesize)
  ifnot (pfileinfo := mem.malloc(file_info_size + handlesize))
    ChangeTempWD
    return 0
  file_data := pfileinfo + file_info_size
  long[pfileinfo][file_info_data] := file_data
  long[pfileinfo][file_info_type] := file_info_fileio
  value := fileio.hopendir(file_data)
  ChangeTempWD
  if value
    mem.free(pfileinfo)
    return 0
  return pfileinfo

PUB readdir(pfileinfo) | file_data
  file_data := long[pfileinfo][file_info_data]
  result := fileio.hreaddir(file_data)

PUB closedir(pfileinfo) | file_data
  file_data := long[pfileinfo][file_info_data]
  result := fileio.hclosedir(file_data)
  mem.free(pfileinfo)

PUB ftell(pfileinfo) | file_data
  file_data := long[pfileinfo][file_info_data]
  result := fileio.htell(file_data)

PUB fseek(pfileinfo, position, origin) | file_data
{{
  This routine moves to the location in the file determined by
  the "newpos" and the "origin".  For origin values of 0, 1 or 2
  the base position is the start of the file, the current position
  or the end of the file respectively.
}}
  file_data := long[pfileinfo][file_info_data]
  if origin == SEEK_CUR
    position += fileio.htell(file_data)
  elseif origin == SEEK_END
    position += fileio.hget_filesize(file_data)
  result := fileio.hseek(file_data, position)

PUB fflush(pfileinfo) | file_data
  file_data := long[pfileinfo][file_info_data]
  result := fileio.hflush(file_data)

PUB mkdir(fname) | pfileinfo, dummy
  result := fileio.mkdir(fname)
{
  pfileinfo := fopen(fname, string("w"))
  ifnot pfileinfo
    return 0
  dummy := 0
  'ser.dbprintf0(string("Write 32 bytes\n"))
  repeat 32/4
    fwrite(@dummy, 1, 4, pfileinfo)
  fclose(pfileinfo)
}

PUB chmod(fname, modebits) | fname1, pathstr[20], filestr[8]
  fname1 := ProcessPath(fname, @pathstr, @filestr)
  result := fileio.chmod(fname1, modebits)
  ChangeTempWD

PUB getmod(fname) | fname1, pathstr[20], filestr[8]
  fname1 := ProcessPath(fname, @pathstr, @filestr)
  result := fileio.getmod(fname1)
  ChangeTempWD

'**** sd and sim versions are identical after this point ****

PUB copyarglist(argc, argv, argv1) | ptr, i, num
  if argc
    ptr := argv1 + (argc<<2) + 4
    repeat i from 0 to argc - 1
      long[argv1][i] := ptr
      num := strsize(long[argv][i]) + 1
      bytemove(ptr, long[argv][i], num)
      ptr += num
    long[argv1][argc]~

PUB KillCaller(argc, argv, argsize, mode) | argv1
  argv1 := mem.malloc(argsize)      ' Allocate space and save args before killing
  copyarglist(argc, argv, argv1)
  kill(8 | (mode & 7))
  result := mem.malloc(argsize)     ' Move args to a new space so we don't fragment memory
  copyarglist(argc, argv1, result)
  mem.free(argv1)

PUB run_stand_alone(argc, argv, argsize, infile, mode) | infile1, argv1, i, dbase, size, cogspi
  ' Determine the file size
  fseek(infile, 0, SEEK_END)
  size := ftell(infile)
  fseek(infile, 0, SEEK_SET)

  ' Get the starting sector number
  result := fileio.hget_first_sector(long[infile][file_info_data])

  ' Open the loader file
  if mode & sys#run_c_program
    infile1 := fopen(string("/bin/_cloader"), string("r"))
  else
    infile1 := fopen(string("/bin/_loader"), string("r"))
  ifnot infile1
    fclose(infile)
    return -1

  ' Copy the arg list to compact it
  i := ((strsize(@currentwd) + 4) >> 2) << 2 ' Get length of CWD + NULL
  argv1 := mem.malloc(argsize + i + 8)
  copyarglist(argc, argv, argv1)

  ' Copy system info after arg list
  long[argv1+argsize] := long[sys#config]
  long[argv1+argsize][1] := long[sys#sd_pins]
  bytemove(argv1+argsize+8, @currentwd, i)
  argsize += i + 8

  ' Stop all the cogs except this one and the SD SPI cog
  cogspi := long[sys#SPI_engine_cog] - 1
  repeat i from 0 to 7
    if i <> cogid and i <> cogspi
      cogstop(i)

  ' Clear and return all the locks
  repeat i from 0 to 7
    lockclr(i)
    lockret(i)

  ' Read the loader into memory
  fread(sys#start, 1, 1000, infile1)
  fclose(infile1)

  ' Add the memory offset to the header
  repeat i from 3 to 7
      word[sys#start][i] += sys#start

  ' Add the parms to the stack
  dbase := word[sys#start][5]
  long[dbase][0]~
  long[dbase][1] := argc
  long[dbase][2] := argv1
  long[dbase][3] := argsize
  long[dbase][4] := result
  long[dbase][5] := size

  ' Start the loader in this cog
  coginit(cogid, $f004, sys#start + 4)

PUB FatalMemoryError
  ser.dbprintf0(string("Not enough memory -- Rebooting!\r\n"))
  waitcnt(8_000_000 + cnt)
  reboot

PUB run(argc, argv, mode) | infile, size, AppAddr, i, cognum, vbase, dbase, argsize, argv1, infile1
  
  if argc < 1
    return -4
    
  infile := fopen(long[argv], string("r"))
  'ser.dbprintf1(string("infile = %8x\n"), infile)
  ifnot infile
    'ser.dbprintf0(string("1\n"))
    return -3

  ' Determine the arg list size
  argsize := 4
  repeat i from 0 to argc - 1
    argsize += 4 + strsize(long[argv][i]) + 1
  argsize := ((argsize + 3) >> 2) << 2 ' Round to nearest long

  if (mode & sys#run_stand_alone)
    return run_stand_alone(argc, argv, argsize, infile, mode)

  ' Check if we need to kill the caller and save the arg list
  if (mode & sys#run_kill_caller)
    argv := KillCaller(argc, argv, argsize, mode)

  ' Determine the file size
  'ser.dbprintf0(string("Determine size\n"))
  {
  fseek(infile, 0, SEEK_END)
  size := ftell(infile)
  fseek(infile, 0, SEEK_SET)
  }
  ' Determine the program size including the VAR section
  AppAddr := mem.malloc(16)
  fread(AppAddr, 1, 16, infile)
  size := word[AppAddr][5] ' size is equal to the value of dbase
  mem.free(AppAddr)
  fseek(infile, 0, SEEK_SET)
  
  'ser.dbprintf1(string("run: size = %d\n"), size)
  ' Allocate the memory plus some for the stack
  if (mode & sys#run_at_address0)
    AppAddr := mem.malloc0(size + 800 + argsize)
    ifnot AppAddr
      fclose(infile)
      if (mode & sys#run_kill_caller)
        FatalMemoryError
      return -2
    fread($6, 1, 6, infile)
    fread($6, 1, size - 6, infile)
    size := AppAddr - 800 - argsize
    AppAddr := 0
    'fclose(infile)
  else
    AppAddr := mem.malloc(size + 800 + argsize)
    'ser.dbprintf2(string("AppAddr = %x, size = %d\n"), AppAddr, size + 800 + argsize)
    ifnot AppAddr
      'ser.str(string("Not enough memory to run "))
      'ser.str(long[argv])
      'ser.tx(13)
      fclose(infile)
      if (mode & sys#run_kill_caller)
        FatalMemoryError
      return -2
    'c.memset(AppAddr, $a5, size + 400)
    fread(AppAddr, 1, size, infile)
    ' Add offset to the data pointers
    repeat i from 3 to 7
      word[AppAddr][i] += AppAddr
    
  fclose(infile)

  'pbase := word[AppAddr][3]
  vbase := word[AppAddr][4]
  dbase := word[AppAddr][5]
  'pcurr := word[AppAddr][6]
  'dcurr := word[AppAddr][7]

  ' Fill the stack with the check word
  longfill(dbase, sys#checkword, 200)

  argv1 := AppAddr + size + 800
  word[dbase][-4] := (@@0) | 2          ' pbase + abort trap and return value
  word[dbase][-3] := @firstvar          ' vbase set to point to firstvar
  word[dbase][-2] := dbase + 64         ' dbase set to a safe place far away
  word[dbase][-1] := word[@@0][2] + @@0 ' Starting address of exit routine -- must be first PUB
  long[dbase]~                          ' Zero result
  long[dbase][1] := argc
  'long[dbase][2] := argv
  long[dbase][2] := argv1

  ' Copy arg list to app
  copyarglist(argc, argv, argv1)
  'ser.dbprintf2(string("After copyarglist -%s- -%s-\n"), long[argv], long[argv1])
  
  if (mode & sys#run_kill_caller)
    mem.free(argv)

  ' Zero the VAR area
  longfill(vbase, 0, (dbase - vbase - 8) >> 2)

  'c.memset(sp, $a5, 100)
  'byte[AppAddr] := 1
  cognum := cognew($f004, AppAddr + 4)
  if cognum == -1
    if (mode & sys#run_kill_caller)
      ser.dbprintf0(string("Out of cogs\r\n"))
    if AppAddr
      mem.free(AppAddr)
    else
      mem.freeraw(16, size + 800 + argsize - 16)
    return -4
  'ser.dbprintf1(string("cfileiosd run: cognum = %d\n"), cognum)

  if (mode & sys#run_kill_caller) or not (mode & sys#run_shell_nowait)
    long[sys#focus] := cognum

  i := ((mode & 7) << 4) | sys#proc_type_spin
  AddToCogTable(cognum, i, long[argv1], AppAddr, size + 800 + argsize)
  'ser.dbprintf0(string("2\n"))

  return cognum

  exit ' Dummy call to exit so it won't be removed

PUB kill(cognum) | entryptr, type, argc, shellstr, focus
  shellstr := string("/bin/shell")
  'ser.dbprintf1(string("Kill cog %d\n"), cognum)
  'ser.dbprintf1(string("cogid = %d\n"), cogid)
  if cognum >> 3
    shellstr~
    cognum &= 7

  entryptr := sys#cogtable + (cognum * 24)
  type := byte[entryptr]
  ifnot type
    'ser.str(string("Process "))
    'ser.dec(cognum)
    'ser.str(string(" is not running", 13))
    return -1
  cogstop(cognum)
  'ser.dbprintf2(string("type = %d, addr = %x\n"), type, word[entryptr+20])
  'ser.dbprintf1(string("name = %s\n"), entryptr+1)
  'if type <> 2 ' Not PASM
  if word[entryptr+22]
    'ser.str(string("kill: free mem at "))
    'ser.hex(word[entryptr+20], 4)
    'ser.tx(13)
    {
    if shellstr
      if strcomp(entryptr+1, shellstr)
        shellstr~
    }
    ifnot word[entryptr+20]
      'ser.str(string("Free boot memory - "))
      'ser.dec(word[entryptr+22] - $10)
      'ser.str(string(" bytes", 13))
      mem.freeraw($10, word[entryptr+22] - $10)
    else
      mem.free(word[entryptr+20])
  byte[entryptr] := 0

  ' Don't start the shell if we were called from run 
  ifnot shellstr
    'ser.str(string("Called from run, don't start shell", 13))
    return

  ' Return if this was not the foreground process
  if cognum <> long[sys#focus]
    return

  ' Attempt to focus on the process that had started this process
  cognum := (type >> 4) & 7
  type := byte[sys#cogtable + (cognum * 24)]
  if type and not (byte[entryptr] & sys#proc_type_driver)
    long[sys#focus] := cognum
    return

  ' Attempt to set the focus on the first available app
  entryptr := sys#cogtable
  repeat cognum from 0 to 7
    type := byte[entryptr]
    if type and not (byte[entryptr] & sys#proc_type_driver)
      long[sys#focus] := cognum
      return

  ' no apps running, start up the shell
  'ser.str(string("App not running", 13))
  run(1, @shellstr, 0)
    

PUB AddToCogTable(cognum, type, name, startaddr, size) | entryptr
  'ser.dbprintf1(string("AddToCogTable %s\n"), name)
  entryptr := sys#cogtable + (cognum * 24)
  byte[entryptr] := type
  bytemove(entryptr+1, name, 19)
  byte[entryptr+19] := 0
  word[entryptr+20] := startaddr
  word[entryptr+22] := size

PUB RemoveFromCogTable(cognum) | entryptr
  entryptr := sys#cogtable + (cognum * 24)
  byte[entryptr] := 0

{{
+-----------------------------------------------------------------------------+
|                       TERMS OF USE: MIT License                             |
+-----------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy |
|of this software and associated documentation files (the "Software"), to deal|
|in the Software without restriction, including without limitation the rights |
|to use, copy, modify, merge, publish, distribute, sublicense, and/or sell    |
|copies of the Software, and to permit persons to whom the Software is        |
|furnished to do so, subject to the following conditions:                     |
|                                                                             |
|The above copyright notice and this permission notice shall be included in   |
|all copies or substantial portions of the Software.                          |
|                                                                             |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR   |
|IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,     |
|FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE  |
|AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER       |
|LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,|
|OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE|
|SOFTWARE.                                                                    |
+-----------------------------------------------------------------------------+
}}
